/* Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.activiti5.spring.autodeployment;

import org.apache.commons.lang3.StringUtils;
import org.springframework.core.io.ByteArrayResource;
import org.springframework.core.io.ContextResource;
import org.springframework.core.io.Resource;

import java.io.IOException;

/**
 * Abstract base class for implementations of {@link AutoDeploymentStrategy}.
 * 
 * @author Tiese Barrell
 */
public abstract class AbstractAutoDeploymentStrategy implements AutoDeploymentStrategy {

  /**
   * Gets the deployment mode this strategy handles.
   * 
   * @return the name of the deployment mode
   */
  protected abstract String getDeploymentMode();

  @Override
  public boolean handlesMode(final String mode) {
    return StringUtils.equalsIgnoreCase(mode, getDeploymentMode());
  }

  /**
   * Determines the name to be used for the provided resource.
   * 
   * @param resource
   *          the resource to get the name for
   * @return the name of the resource
   */
  protected String determineResourceName(final Resource resource) {
    String resourceName = null;

    if (resource instanceof ContextResource) {
      resourceName = ((ContextResource) resource).getPathWithinContext();

    } else if (resource instanceof ByteArrayResource) {
      resourceName = resource.getDescription();

    } else {
      try {
        resourceName = resource.getFile().getAbsolutePath();
      } catch (IOException e) {
        resourceName = resource.getFilename();
      }
    }
    return resourceName;
  }

}
